/*
 * (c) Copyright 1992 by Panagiotis Tsirigotis
 * All rights reserved.  The file named COPYRIGHT specifies the terms 
 * and conditions for redistribution.
 */

#ifndef FLAGS_H
#define FLAGS_H

/*
 * $Id: flags.h,v 6.2 1993/04/22 23:37:35 panos Exp $
 */

/*
 * Flags
 */
#define SOFT_RECONFIG_FLAG			1
#define HARD_RECONFIG_FLAG			2
#define TERMINATE_FLAG				3
#define DUMP_FLAG						4
#define RETRY_FLAG					5
#define CONSISTENCY_FLAG			6
#define CHILD_FLAG					7
#define QUIT_FLAG						8
#define PERIODIC_CHECK_FLAG		9

#endif	/* FLAGS_H */
